# from click import command, pass_obj

# from ..util.cli import DreemCommandName, dreem_command


# @command(DreemCommandName.TEST.value)
# # Pass context object
# @pass_obj
# # Turn into DREEM command
# @dreem_command()
# def cli():
#     return run()


# def run():
#     print("Running tests")
